/*
 * This file is part of qtutil
 * Copyright (c) 2000 David Deaven (deaven@execpc.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Author: deaven $
 * $Id: QTAtomImpl.h,v 1.2 2000/10/14 04:47:06 deaven Exp $
 */

#ifndef __QTAtomImpl__
#define __QTAtomImpl__

#include <stdio.h>
#include <time.h>
#include <string.h>

typedef unsigned int int32;
typedef unsigned short int16;
typedef float fixed32;
typedef unsigned char byte;

// ================================================================
class QTAtomImpl {
public:
    QTAtomImpl(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	Size(size),
	HeaderSize(headerSize),
	Offset(offset),
	FilePtr(parent->filePtr())
	{
	    bcopy(type, Type, 4);
	}
    QTAtomImpl(FILE* fptr) :
	Size(0),
	HeaderSize(0),
	Offset(0),
	FilePtr(fptr)
	{
	    bcopy("FILE", Type, 4);
	}

    virtual ~QTAtomImpl() { }
    virtual QTAtomImpl** children(int* count);
    virtual void print();

    FILE* filePtr() { return FilePtr; }
    unsigned long read_buffer(void* start, void* end, unsigned long atom_offset = 0);

    char Type[4];
    unsigned long Size;
    unsigned long HeaderSize;
    unsigned long Offset;

    static int cmp32(const void* a, const void* b) {
	byte* ba = (byte*) a;
	byte* bb = (byte*) b;
	if (ba == bb) return 1;
	return (ba && bb && 
	    ba[0] == bb[0] &&
	    ba[1] == bb[1] &&
	    ba[2] == bb[2] &&
	    ba[3] == bb[3]) ? 1 : 0;
    }

    static time_t convert_mac_time(int32 mac_time);

protected:
    QTAtomImpl* read_atom_at(unsigned long offset, unsigned long max_size);
    static void convert_int16(int16* data);
    static void convert_int32(int32* data);
    static int read_int32(FILE* f, int32* val);
    static void convert_fixed32(fixed32* data);
    static void print_mac_time(int32 mac_time);

private:
    FILE* FilePtr;
};

// ================================================================
class QTmvhd : public QTAtomImpl {
public:
    QTmvhd(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTAtomImpl(parent, type, size, headerSize, offset)
    {
	int i;
	read_buffer(&(this->version), &(this->eos));
	convert_int32(&create_time);
	convert_int32(&modify_time);
	convert_int32(&time_scale);
	convert_int32(&duration);
	convert_int32(&preferred_rate);
	for (i = 0; i < 9; i++)
	{
	    convert_fixed32(&(matrix[i]));
	}
	convert_int32(&preview_time);
	convert_int32(&preview_duration);
	convert_int32(&poster_time);
	convert_int32(&selection_time);
	convert_int32(&selection_duration);
	convert_int32(&current_time);
	convert_int32(&next_track_id);
    }

    byte version;
    byte flags[3];
    int32 create_time;
    int32 modify_time;
    int32 time_scale;
    int32 duration;
    int32 preferred_rate;
    byte volume[2];
    byte reserved[10];
    fixed32 matrix[9];
    int32 preview_time;
    int32 preview_duration;
    int32 poster_time;
    int32 selection_time;
    int32 selection_duration;
    int32 current_time;
    int32 next_track_id;
    int32 eos;
    
    void print()
    {
	QTAtomImpl::print();
	printf(" [create_time:");
	print_mac_time(create_time);
	printf(" modify_time:");
	print_mac_time(modify_time);
	printf(" time_scale:%d duration:%d next_track_id:%d]", 
	    (int) time_scale, (int) duration, (int) next_track_id);
    }
};



// ================================================================
class QTtkhd : public QTAtomImpl {
public:
    byte version;
    byte flags[3];
    int32 create_time;
    int32 modify_time;
    int32 track_id;
    byte reserved0[4];
    int32 duration;
    byte reserved1[8];
    byte layer[2];
    byte alternate_group[2];
    byte volume[2];
    byte reserved2[2];
    fixed32 matrix[9];
    fixed32 track_width;
    fixed32 track_height;
    int32 eos;

    QTtkhd(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTAtomImpl(parent, type, size, headerSize, offset)
    {
	int i;
	read_buffer(&(this->version), &(this->eos));
	convert_int32(&create_time);
	convert_int32(&modify_time);
	convert_int32(&track_id);
	convert_int32(&duration);
	for (i = 0; i < 9; i++)
	{
	    convert_fixed32(&(matrix[i]));
	}
	convert_fixed32(&track_width);
	convert_fixed32(&track_height);
    }

    enum {
	TK_ENABLED = 0x0001,
	TK_IN_MOVIE = 0x0002,
	TK_IN_PREVIEW = 0x0004,
	TK_IN_POSTER = 0x0008
    };

    void print()
    {
	QTAtomImpl::print();
	printf(" [%dx%d create_time:",
	    (int) track_width, (int) track_height);
	print_mac_time(create_time);
	printf(" modify_time:");
	print_mac_time(modify_time);
	char buf[128];
	buf[0] = '\0';
	if (flags[0] & TK_ENABLED) strcat(buf, "enabled ");
	if (flags[0] & TK_IN_MOVIE) strcat(buf, "movie ");
	if (flags[0] & TK_IN_PREVIEW) strcat(buf, "preview ");
	if (flags[0] & TK_IN_POSTER) strcat(buf, "poster ");

	printf(" duration:%d flags:\"%s\"]", 
	    (int) duration, buf);
    }
};



// ================================================================
class QThdlr : public QTAtomImpl {
public:
    byte version;
    byte flags[3];
    byte type[4];
    byte subtype[4];
    int32 manufacturer;
    int32 cflags;
    int32 cflags_mask;
    int32 eos;

    QThdlr(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTAtomImpl(parent, type, size, headerSize, offset)
    {
	read_buffer(&(this->version), &(this->eos));
	convert_int32(&manufacturer);
	convert_int32(&cflags);
	convert_int32(&cflags_mask);
    }

    void print()
    {
	QTAtomImpl::print();
	printf(" [type:%c%c%c%c subtype:%c%c%c%c]",
	    type[0], type[1], type[2], type[3],
	    subtype[0], subtype[1], subtype[2], subtype[3]);
    }
};

// ================================================================
class QTstsz : public QTAtomImpl {
public:
    byte version;
    byte flags[3];
    int32 sample_size;
    int32 count;
    int32 eos;

    int32* sizes;

    QTstsz(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTAtomImpl(parent, type, size, headerSize, offset)
    {
	unsigned long atom_offset = read_buffer(&(this->version), &(this->eos));
	convert_int32(&count);
	convert_int32(&sample_size);

	if (sample_size == 0)
	{
	    sizes = new int32[count];
	    byte* start = (byte*) &(sizes[0]);
	    byte* end = start + 4 * count;
	    atom_offset += read_buffer(start, end, atom_offset);
	    int32 i;
	    for (i = 0; i < count; i++)
	    {
		convert_int32(&(sizes[i]));
	    }
	}
	else
	{
	    sizes = NULL;
	}
    }

    void print()
    {
	QTAtomImpl::print();
	if (sample_size != 0)
	{
	    printf(" [%d samples size=%d]", count, sample_size);
	}
	else
	{
	    printf(" [%d samples size={", count);
	    int32 i;
	    for (i = 0; i < count; i++)
	    {
		printf("%s%d", i > 0 ? "," : "", sizes[i]);
	    }
	    printf("}]");
	}
    }
};

// ================================================================
class QTTableAtomImpl : public QTAtomImpl {
public:
    byte version;
    byte flags[3];
    int32 count;
    int32 eos;

    QTTableAtomImpl(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTAtomImpl(parent, type, size, headerSize, offset)
    {
	InitialAtomOffset = read_buffer(&(this->version), &(this->eos));
	convert_int32(&count);
    }

    void print()
    {
	QTAtomImpl::print();
	printf(" [%d items={", count);
	int32 i;
	for (i = 0; i < count; i++)
	{
	    if (i > 0) printf(",");
	    printItem(i);
	}
	printf("}]");
    }

    virtual void printItem(int i) = 0;

    unsigned long initialAtomOffset() const { return InitialAtomOffset; }

private:
    unsigned long InitialAtomOffset;

};

// ================================================================
class QTdref : public QTTableAtomImpl {
public:

    struct reference {
	int32 size;
	byte type[4];
	byte version;
	byte flags[3];
	byte* data;
    };

    struct reference* references;

    QTdref(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTTableAtomImpl(parent, type, size, headerSize, offset)
    {
	unsigned long atom_offset = initialAtomOffset();
	references = new reference[count];
	int32 i;
	for (i = 0; i < count; i++)
	{
	    byte* start = (byte*) &(references[i]);
	    byte* end = start + 12;
	    atom_offset += read_buffer(start, end, atom_offset);
	    convert_int32(&(references[i].size));
	    references[i].data = new byte[references[i].size];
	    start = references[i].data;
	    end = start + references[i].size - 12;
	    atom_offset += read_buffer(start, end, atom_offset);
	}
    }

    void printItem(int i)
    {
	printf("%c%c%c%c@%db",
	    references[i].type[0], references[i].type[1],
	    references[i].type[2], references[i].type[3],
	    references[i].size);
    }
};



// ================================================================
class QTstsd : public QTTableAtomImpl {
public:
    
    struct sample {
	int32 size;
	byte format[4];
	byte reserved[6];
	int16 reference;
	byte* data;
    };

    struct sample* samples;

    QTstsd(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTTableAtomImpl(parent, type, size, headerSize, offset)
    {
	unsigned long atom_offset = initialAtomOffset();
	samples = new sample[count];
	int32 i;
	for (i = 0; i < count; i++)
	{
	    byte* start = (byte*) &(samples[i]);
	    byte* end = start + 16;
	    atom_offset += read_buffer(start, end, atom_offset);
	    convert_int32(&(samples[i].size));
	    convert_int16(&(samples[i].reference));
	    samples[i].data = new byte[samples[i].size];
	    start = samples[i].data;
	    end = start + samples[i].size - 16;
	    atom_offset += read_buffer(start, end, atom_offset);
	}
    }

    void printItem(int i)
    {
	printf("%c%c%c%c@%db",
	    samples[i].format[0], samples[i].format[1],
	    samples[i].format[2], samples[i].format[3],
	    samples[i].size);
    }
};

// ================================================================
class QTstsc : public QTTableAtomImpl {
public:

    struct sample_to_chunk_entry {
	int32 first_chunk;
	int32 samples_per_chunk;
	int32 sample_desc_id;
    };

    struct sample_to_chunk_entry* sample_to_chunk;

    QTstsc(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTTableAtomImpl(parent, type, size, headerSize, offset)
    {
	unsigned long atom_offset = initialAtomOffset();
	sample_to_chunk = new sample_to_chunk_entry[count];
	int32 i;
	for (i = 0; i < count; i++)
	{
	    byte* start = (byte*) &(sample_to_chunk[i]);
	    byte* end = start + 12;
	    atom_offset += read_buffer(start, end, atom_offset);
	    convert_int32(&(sample_to_chunk[i].first_chunk));
	    convert_int32(&(sample_to_chunk[i].samples_per_chunk));
	    convert_int32(&(sample_to_chunk[i].sample_desc_id));
	}
    }

    void printItem(int i)
    {
	printf("[first:0x%x,spc:%d,desc_id:%d]",
	    sample_to_chunk[i].first_chunk,
	    sample_to_chunk[i].samples_per_chunk,
	    sample_to_chunk[i].sample_desc_id);
    }
};

// ================================================================
class QTstco : public QTTableAtomImpl {
public:

    int32* chunk_offsets;

    QTstco(QTAtomImpl* parent, char type[4], unsigned long size,
	unsigned long headerSize, unsigned long offset) :
	QTTableAtomImpl(parent, type, size, headerSize, offset)
    {
	unsigned long atom_offset = initialAtomOffset();
	chunk_offsets = new int32[count];
	byte* start = (byte*) &(chunk_offsets[0]);
	byte* end = start + 4 * count;
	atom_offset += read_buffer(start, end, atom_offset);
	int32 i;
	for (i = 0; i < count; i++)
	{
	    convert_int32(&(chunk_offsets[i]));
	}
    }

    void printItem(int i)
    {
	printf("offset:0x%x", chunk_offsets[i]);
    }
};

#endif // __QTAtomImpl__
